{/ checkCRC.pas
//
// CRC32 checking iterators.
//
// Copyright (c)1993 by Ark Belousov
// All Rights Reserved.
//
// V1.0, 21.10.93 - first release.
/}

unit checkCRC;
{$i options.off}
interface

uses Types;

const allOne  = $FFFFFFFF;
 CRC32start = allOne;

type CRCiterator = object
     constructor init;
     destructor done;   virtual;
     procedure checkByte(b :byte); virtual;
     procedure checkBuffer(var buf; size :word);
      virtual;
 end {CRCiterator};

 CRC32iterator = object(CRCiterator)
     crc :long;
     constructor init;
     function get :long;  virtual;
     procedure checkByte(b :byte); virtual;
     procedure checkBuffer(var buf; size :word);
      virtual;
 end {CRC32iterator};

function checkFileCRC32(fname :string; var IOres :word) :long;


{-------------------------} implementation {-------------------------}

const CRC32Polynomial = $EDB88320;
 optimumBuffer = 1 shl 15 {32768};

var CRCtable :array[byte] of long;

{//++++++++++++++++++++++++++++++++++++++++++++++}
constructor CRCiterator.init;
begin
end {CRCiterator.init};

{//++++++++++++++++++++++++++++++++++++++++++++++}
destructor CRCiterator.done;
begin
end {CRCiterator.done};

{//++++++++++++++++++++++++++++++++++++++++++++++}
procedure CRCiterator.checkByte(b :byte);
begin runError(211)
end {CRCiterator.checkByte};

{//++++++++++++++++++++++++++++++++++++++++++++++}
procedure CRCiterator.checkBuffer(var buf; size :word);
var p :word;
 buffer :sequence absolute buf;

begin p := 0;
 while(p < size) do begin
  checkByte(buffer[p]); INC(p);
 end {while};
end {CRCiterator.checkBuffer};

{//++++++++++++++++++++++++++++++++++++++++++++++}
constructor CRC32iterator.init;
begin crc := CRC32start;
end {CRC32iterator.init};

{//++++++++++++++++++++++++++++++++++++++++++++++}
function CRC32iterator.get :long;
begin get := not crc;
end {CRC32iterator.get};

{//++++++++++++++++++++++++++++++++++++++++++++++
// Routine calculates the CRC for a block of data using table
// lookup method. It accepts an original value for the CRC,
// and returns the updated value.
//++++++++++++++++++++++++++++++++++++++++++++++}
procedure CRC32iterator.checkByte(b :byte);
begin { Warning: long are signed, but SHR are unsigned operation }
{!!! crc := (crc shr 8) xor CRCtable[lo(crc) xor b]; !!!}
asm {------ es:di := @crc }
 les di, self; lea di, CRC32iterator[es:di].crc
 {------ dx:ax := crc }
 mov ax, es:[di]; mov dx, es:[di] + 2
 {------ bx := [al=lo(crc) xor b] }
 mov bl, al; xor bh, bh
 xor bl, &b; shl bx,1; shl bx,1
 {------ [es:di] := (dx:ax >> 8) ^ CRCtable[bx] }
 mov al, dh; mov ah, dl
 xor ax, word(CRCtable[bx])
 cld; stosw
 mov ax, word(CRCtable[bx] + 2)
 xor al, dh
 stosw
end {asm};
end {CRC32iterator.checkByte};

{//++++++++++++++++++++++++++++++++++++++++++++++
// Routine redeclared and inlined for speeding.
//++++++++++++++++++++++++++++++++++++++++++++++}
procedure CRC32iterator.checkBuffer(var buf; size :word);
begin if(size = 0) then exit;
 asm
 {------ es := ds = seg(CRCtable) }
 mov ax, ds; mov es, ax
 {------ dx:cx := crc }
 lds di, self
 mov cx, word(CRC32iterator[di].crc)
 mov dx, word(CRC32iterator[di].crc + 2)
 {------ ds:si := @buf }
 lds si, &buf; cld
 {------ dx:cx = crc; ds:si = buf[i]; es = seg(CRCtable) }
 end {asm};

 repeat
(* while(size > 0) do begin*)
{!!!  crc := (crc shr 8) xor CRCtable[lo(crc) xor buf[i]]; !!!}
(*  DEC(size);*)
  asm
  {------ bx := [cl=lo(crc) xor buf[i++]] }
  mov bl, cl; xor bh, bh
  lodsb; xor bl, al
  shl bx,1; shl bx,1
  {------ dx:cx := (dx:cx >> 8) ^ es:CRCtable[bx] }
  mov cl, ch; mov ch, dl
  xor cx, word(es:CRCtable[bx])

  mov ax, word(es:CRCtable[bx] + 2)
  xor al, dh; mov dx, ax
  end {asm};
  DEC(size);
 until(size = 0);
(* end {while};*)

 asm
 {------ crc := dx:cx }
 lds di, self
 mov word(CRC32iterator[di].crc), cx
 mov word(CRC32iterator[di].crc + 2), dx
 {------ ds := es }
 mov ax, es; mov ds, ax
 end {asm};
end {CRC32iterator.checkBuffer};

{//++++++++++++++++++++++++++++++++++++++++++++++
// This routine is responsible for performing the calculation
// of the 32-bit CRC for the entire file. We precondition the
// CRC value with all 1's, then invert every bit after the
// entire file has been done. This gives us a CRC value that
// corrresponds with the values calculated by PKZIP and ARJ.
//++++++++++++++++++++++++++++++++++++++++++++++}
var buf :array[1..optimumBuffer] of byte;

function checkFileCRC32(fname :string; var IOres :word) :long;
var f :file;
 crc32 :CRC32iterator;
 cnt :word;
 fm :byte;

begin fm := fileMode; fileMode := 0;
 assign(f, fname); reset(f, 1);
 IOres := IOresult; fileMode := fm;
 if(IOres <> 0) then
  checkFileCRC32 := 0
 else begin
  crc32.init;
  repeat blockRead(f, buf, sizeOf(buf), cnt);
   crc32.checkBuffer(buf, cnt);
  until(cnt = 0);
  checkFileCRC32 := crc32.get;
  IOres := IOresult;
 end {if};
 close(f);
end {checkFileCRC32};

{-----------------------------------------------------------
  Instead of performing a straightforward calculation of the
  32-bit CRC using a series of logical operations, program
  uses faster table lookup method.
------------------------------------------------------------}
var i, j :int;
 crc :long;

begin for i := minByte to maxByte do begin
     crc := i;
     for j := 1 to bitsPerByte do begin
  if(odd(crc)) then
{ Warning: long are signed, but SHR are unsigned operation }
   crc := (crc shr 1) xor CRC32Polynomial
  else crc := crc shr 1;
     end {for};
     CRCtable[ i ] := crc;
 end {for};
end {checkCRC}.

